// This code was generated by a tool.
// https://github.com/mezoni/parser_builder

/// Parses the CSV data and returns the result as a `List<List<String>>`.
/// - Will not parse numbers
/// - The character `,` is used as a field separator
/// - Line endings are `\n`, `\r\n` or `\r`
/// - The start and end of strings is the character `"`
/// - Escaping a character `"` in a string is parsed via sequence `""`
/// - Exception `FormatException` will be thrown if parsing fails
List<List<String>> parse(String source) {
  final state = State(source);
  final result = _parse(state);
  if (!state.ok) {
    final message = _errorMessage(source, state.errors);
    throw FormatException('\n$message');
  }

  return result!;
}

void _ws(State<String> state) {
  final source = state.source;
  while (state.pos < source.length) {
    final c = source.codeUnitAt(state.pos);
    final ok = c <= 9 ? c == 9 : c == 32;
    if (!ok) {
      break;
    }
    state.pos++;
  }
  state.ok = true;
}

void _quote(State<String> state) {
  final source = state.source;
  state.ok = state.pos < source.length && source.codeUnitAt(state.pos) == 34;
  if (state.ok) {
    state.pos += 1;
  } else {
    state.fail(state.pos, ParseError.expected, '"');
  }
}

void _openQuote(State<String> state) {
  final $pos = state.pos;
  _ws(state);
  if (state.ok) {
    _quote(state);
    if (!state.ok) {
      state.pos = $pos;
    }
  }
}

List<int>? _chars(State<String> state) {
  List<int>? $0;
  final source = state.source;
  final $list = <int>[];
  while (true) {
    int? $1;
    state.ok = state.pos < source.length;
    if (state.ok) {
      final pos = state.pos;
      final c = source.readRune(state);
      state.ok = c != 34;
      if (state.ok) {
        $1 = c;
      } else {
        state.pos = pos;
        state.fail(state.pos, ParseError.character, null);
      }
    } else {
      state.fail(state.pos, ParseError.character, null);
    }
    if (!state.ok) {
      state.ok = state.pos + 1 < source.length &&
          source.codeUnitAt(state.pos) == 34 &&
          source.codeUnitAt(state.pos + 1) == 34;
      if (state.ok) {
        state.pos += 2;
      } else {
        state.fail(state.pos, ParseError.expected, '""');
      }
      if (state.ok) {
        $1 = 34;
      }
    }
    if (!state.ok) {
      break;
    }
    $list.add($1!);
  }
  state.ok = true;
  if (state.ok) {
    $0 = $list;
  }
  return $0;
}

void _closeQuote(State<String> state) {
  final $pos = state.pos;
  _quote(state);
  if (state.ok) {
    _ws(state);
    if (!state.ok) {
      state.pos = $pos;
    }
  }
}

String? _string(State<String> state) {
  String? $0;
  final $pos = state.start;
  state.start = state.pos;
  final $pos1 = state.setLastErrorPos(-1);
  final $pos2 = state.pos;
  _openQuote(state);
  if (state.ok) {
    List<int>? $1;
    $1 = _chars(state);
    if (state.ok) {
      _closeQuote(state);
      if (!state.ok) {
        state.fail(
            state.lastErrorPos, ParseError.message, 'Unterminated string',
            start: state.start);
      }
      if (state.ok) {
        final v1 = $1!;
        $0 = String.fromCharCodes(v1);
      }
    }
  }
  if (!state.ok) {
    state.pos = $pos2;
  }
  state.restoreLastErrorPos($pos1);
  state.start = $pos;
  return $0;
}

String? _field(State<String> state) {
  String? $0;
  final source = state.source;
  $0 = _string(state);
  if (!state.ok) {
    final $pos = state.pos;
    while (state.pos < source.length) {
      final pos = state.pos;
      final c = source.readRune(state);
      final ok = !(c <= 13
          ? c <= 10
              ? c == 10
              : c == 13
          : c <= 34
              ? c == 34
              : c == 44);
      if (!ok) {
        state.pos = pos;
        break;
      }
    }
    state.ok = true;
    if (state.ok) {
      $0 = $pos == state.pos ? '' : source.substring($pos, state.pos);
    }
  }
  return $0;
}

List<String>? _row(State<String> state) {
  List<String>? $0;
  final source = state.source;
  var $pos = state.pos;
  final $list = <String>[];
  while (true) {
    String? $1;
    $1 = _field(state);
    if (!state.ok) {
      state.pos = $pos;
      break;
    }
    $list.add($1!);
    $pos = state.pos;
    state.ok = state.pos < source.length && source.codeUnitAt(state.pos) == 44;
    if (state.ok) {
      state.pos += 1;
    } else {
      state.fail(state.pos, ParseError.expected, ',');
    }
    if (!state.ok) {
      break;
    }
  }
  state.ok = $list.isNotEmpty;
  if (state.ok) {
    $0 = $list;
  }
  return $0;
}

void _eol(State<String> state) {
  final source = state.source;
  state.ok = state.pos < source.length;
  if (state.ok) {
    final pos = state.pos;
    final c = source.codeUnitAt(pos);
    state.ok = false;
    if (c == 10) {
      state.ok = true;
      state.pos += 1;
    } else if (c == 13) {
      if (source.startsWith('\r\n', pos)) {
        state.ok = true;
        state.pos += 2;
      } else {
        state.ok = true;
        state.pos += 1;
      }
    }
  }
  if (!state.ok) {
    state.fail(state.pos, ParseError.expected, '\n');
    state.fail(state.pos, ParseError.expected, '\r\n');
    state.fail(state.pos, ParseError.expected, '\r');
  }
}

void _eof(State<String> state) {
  final source = state.source;
  state.ok = state.pos >= source.length;
  if (!state.ok) {
    state.fail(state.pos, ParseError.expected, 'EOF');
  }
}

void _rowEnding(State<String> state) {
  final $pos = state.pos;
  _eol(state);
  if (state.ok) {
    final $pos1 = state.pos;
    final $log = state.log;
    state.log = false;
    _eof(state);
    state.log = $log;
    state.ok = !state.ok;
    if (!state.ok) {
      state.pos = $pos1;
      state.fail(state.pos, ParseError.message, 'Unknown error');
    }
    if (!state.ok) {
      state.pos = $pos;
    }
  }
}

List<List<String>>? _rows(State<String> state) {
  List<List<String>>? $0;
  final $pos = state.pos;
  var $pos1 = state.pos;
  final $list = <List<String>>[];
  while (true) {
    List<String>? $1;
    $1 = _row(state);
    if (!state.ok) {
      state.pos = $pos1;
      break;
    }
    $list.add($1!);
    $pos1 = state.pos;
    _rowEnding(state);
    if (!state.ok) {
      break;
    }
  }
  state.ok = $list.isNotEmpty;
  if (state.ok) {
    $0 = $list;
  }
  if (state.ok) {
    _eol(state);
    if (!state.ok) {
      state.ok = true;
    }
    if (!state.ok) {
      $0 = null;
      state.pos = $pos;
    }
  }
  return $0;
}

List<List<String>>? _parse(State<String> state) {
  List<List<String>>? $0;
  final $pos = state.pos;
  $0 = _rows(state);
  if (state.ok) {
    _eof(state);
    if (!state.ok) {
      $0 = null;
      state.pos = $pos;
    }
  }
  return $0;
}

String _errorMessage(String source, List<ParseError> errors) {
  final sb = StringBuffer();
  for (var i = 0; i < errors.length; i++) {
    if (sb.isNotEmpty) {
      sb.writeln();
      sb.writeln();
    }

    final error = errors[i];
    final start = error.start;
    final end = error.end;
    RangeError.checkValidRange(start, end, source.length);
    var row = 1;
    var lineStart = 0, next = 0, pos = 0;
    while (pos < source.length) {
      final c = source.codeUnitAt(pos++);
      if (c == 0xa || c == 0xd) {
        next = c == 0xa ? 0xd : 0xa;
        if (pos < source.length && source.codeUnitAt(pos) == next) {
          pos++;
        }

        if (pos - 1 >= start) {
          break;
        }

        row++;
        lineStart = pos;
      }
    }

    int max(int x, int y) => x > y ? x : y;
    int min(int x, int y) => x < y ? x : y;
    final sourceLen = source.length;
    final lineLimit = min(80, sourceLen);
    final start2 = start;
    final end2 = min(start2 + lineLimit, end);
    final errorLen = end2 - start;
    final extraLen = lineLimit - errorLen;
    final rightLen = min(sourceLen - end2, extraLen - (extraLen >> 1));
    final leftLen = min(start, max(0, lineLimit - errorLen - rightLen));
    final list = <int>[];
    final iterator = RuneIterator.at(source, start2);
    for (var i = 0; i < leftLen; i++) {
      if (!iterator.movePrevious()) {
        break;
      }

      list.add(iterator.current);
    }

    final column = start - lineStart + 1;
    final left = String.fromCharCodes(list.reversed);
    final end3 = min(sourceLen, start2 + (lineLimit - leftLen));
    final indicatorLen = max(1, errorLen);
    final right = source.substring(start2, end3);
    var text = left + right;
    text = text.replaceAll('\n', ' ');
    text = text.replaceAll('\r', ' ');
    text = text.replaceAll('\t', ' ');
    sb.writeln('line $row, column $column: $error');
    sb.writeln(text);
    sb.write(' ' * leftLen + '^' * indicatorLen);
  }

  return sb.toString();
}

extension on String {
  @pragma('vm:prefer-inline')
  // ignore: unused_element
  int readRune(State<String> state) {
    final w1 = codeUnitAt(state.pos++);
    if (w1 > 0xd7ff && w1 < 0xe000) {
      if (state.pos < length) {
        final w2 = codeUnitAt(state.pos++);
        if ((w2 & 0xfc00) == 0xdc00) {
          return 0x10000 + ((w1 & 0x3ff) << 10) + (w2 & 0x3ff);
        }

        state.pos--;
      }

      throw FormatException('Invalid UTF-16 character', this, state.pos - 1);
    }

    return w1;
  }

  @pragma('vm:prefer-inline')
  // ignore: unused_element
  int runeAt(int index) {
    final w1 = codeUnitAt(index++);
    if (w1 > 0xd7ff && w1 < 0xe000) {
      if (index < length) {
        final w2 = codeUnitAt(index);
        if ((w2 & 0xfc00) == 0xdc00) {
          return 0x10000 + ((w1 & 0x3ff) << 10) + (w2 & 0x3ff);
        }
      }

      throw FormatException('Invalid UTF-16 character', this, index - 1);
    }

    return w1;
  }

  /// Returns a slice (substring) of the string from [start] to [end].
  @pragma('vm:prefer-inline')
  // ignore: unused_element
  String slice(int start, int end) {
    return substring(start, end);
  }
}

class ParseError {
  static const character = 0;

  static const expected = 1;

  static const message = 2;

  static const unexpected = 3;

  final int end;

  final int start;

  final String text;

  ParseError(this.start, this.end, this.text);

  @override
  int get hashCode => end.hashCode ^ start.hashCode ^ text.hashCode;

  @override
  bool operator ==(other) =>
      other is ParseError &&
      other.end == end &&
      other.start == start &&
      other.text == text;

  @override
  String toString() {
    return text;
  }
}

class State<T> {
  dynamic context;

  int errorPos = -1;

  int lastErrorPos = -1;

  int minErrorPos = -1;

  bool log = true;

  bool ok = false;

  int pos = 0;

  int start = 0;

  final T source;

  final List<int> _kinds = List.filled(150, 0);

  int _length = 0;

  final List<int> _lengths = List.filled(150, 0);

  final List<int> _starts = List.filled(150, 0);

  final List<Object?> _values = List.filled(150, null);

  State(this.source);

  List<ParseError> get errors => _buildErrors();

  @pragma('vm:prefer-inline')
  void fail(int pos, int kind, Object? value,
      {int length = -1, int start = -1}) {
    ok = false;
    if (log) {
      if (errorPos <= pos && minErrorPos <= pos) {
        if (errorPos < pos) {
          errorPos = pos;
          _length = 0;
        }

        _kinds[_length] = kind;
        _lengths[_length] = length;
        _starts[_length] = start;
        _values[_length] = value;
        _length++;
      }

      if (lastErrorPos < pos) {
        lastErrorPos = pos;
      }
    }
  }

  @pragma('vm:prefer-inline')
  void restoreLastErrorPos(int pos) {
    if (lastErrorPos < pos) {
      lastErrorPos = pos;
    }
  }

  @pragma('vm:prefer-inline')
  int setLastErrorPos(int pos) {
    final result = lastErrorPos;
    lastErrorPos = pos;
    return result;
  }

  @override
  String toString() {
    if (source is String) {
      final s = source as String;
      if (pos >= s.length) {
        return '$pos:';
      }

      var length = s.length - pos;
      length = length > 40 ? 40 : length;
      final string = s.substring(pos, pos + length);
      return '$pos:$string';
    } else {
      return super.toString();
    }
  }

  List<ParseError> _buildErrors() {
    int max(int x, int y) => x > y ? x : y;

    int min(int x, int y) => x < y ? x : y;

    int getStart(int index) {
      var start = _starts[index];
      if (start < 0) {
        start = errorPos;
      }

      start = min(start, errorPos);
      return start;
    }

    int getEnd(int index) {
      start = getStart(index);
      var end = _starts[index];
      if (end < 0) {
        end = errorPos;
      }

      end = max(end, errorPos);
      end = max(end, start);
      start = min(start, end);
      final length = _lengths[index];
      if (length >= 0) {
        end = start + length;
      }

      return end;
    }

    final result = <ParseError>[];
    final expected = <int, List>{};
    for (var i = 0; i < _length; i++) {
      final kind = _kinds[i];
      if (kind == ParseError.expected) {
        final value = _values[i];
        final start = getStart(i);
        var list = expected[start];
        if (list == null) {
          list = [];
          expected[start] = list;
        }

        list.add(value);
      }
    }

    for (final start in expected.keys) {
      final values = expected[start]!.toSet().map((e) => _escape(e));
      final text = 'Expected: ${values.join(', ')}';
      final error = ParseError(start, start, text);
      result.add(error);
    }

    for (var i = 0; i < _length; i++) {
      final start = getStart(i);
      final end = getEnd(i);
      final value = _values[i];
      final kind = _kinds[i];
      switch (kind) {
        case ParseError.character:
          if (source is String) {
            final string = source as String;
            if (start < string.length) {
              final value = string.runeAt(errorPos);
              final length = value >= 0xffff ? 2 : 1;
              final escaped = _escape(value);
              final error =
                  ParseError(start, start + length, 'Unexpected $escaped');
              result.add(error);
            } else {
              final error = ParseError(start, start, "Unexpected 'EOF'");
              result.add(error);
            }
          } else {
            final error = ParseError(start, start, 'Unexpected character');
            result.add(error);
          }

          break;
        case ParseError.expected:
          break;
        case ParseError.message:
          final error = ParseError(start, end, '$value');
          result.add(error);
          break;
        case ParseError.unexpected:
          final escaped = _escape(value);
          final error = ParseError(start, end, 'Unexpected $escaped');
          result.add(error);
          break;
        default:
          final error = ParseError(start, end, '$value');
          result.add(error);
      }
    }

    return result.toSet().toList();
  }

  String _escape(Object? value, [bool quote = true]) {
    if (value is int) {
      if (value >= 0 && value <= 0xd7ff ||
          value >= 0xe000 && value <= 0x10ffff) {
        value = String.fromCharCode(value);
      } else {
        return value.toString();
      }
    } else if (value is! String) {
      return value.toString();
    }

    final map = {
      '\b': '\\b',
      '\f': '\\f',
      '\n': '\\n',
      '\r': '\\r',
      '\t': '\\t',
      '\v': '\\v',
    };
    var result = value.toString();
    for (final key in map.keys) {
      result = result.replaceAll(key, map[key]!);
    }

    if (quote) {
      result = "'$result'";
    }

    return result;
  }
}
